# https://github.com/oobabooga/one-click-installers

import requests
import os
import subprocess
import zipfile
import sys

script_dir = os.getcwd()

def run_cmd(cmd, assert_success=False, environment=False, capture_output=False, env=None):
    # Use the conda environment
    if environment:
        conda_env_path = os.path.join(script_dir, "installer_files", "env")
        if sys.platform.startswith("win"):
            conda_bat_path = os.path.join(script_dir, "installer_files", "conda", "condabin", "conda.bat")
            cmd = "\"" + conda_bat_path + "\" activate \"" + conda_env_path + "\" >nul && " + cmd
        else:
            conda_sh_path = os.path.join(script_dir, "installer_files", "conda", "etc", "profile.d", "conda.sh")
            cmd = ". \"" + conda_sh_path + "\" && conda activate \"" + conda_env_path + "\" && " + cmd

    # Run shell commands
    result = subprocess.run(cmd, shell=True, capture_output=capture_output, env=env)

    # Assert the command ran successfully
    if assert_success and result.returncode != 0:
        print("Command '" + cmd + "' failed with exit status code '" + str(result.returncode) + "'. Exiting...")
        sys.exit()
    return result


def check_env():
    # If we have access to conda, we are probably in an environment
    conda_exist = run_cmd("conda", environment=True, capture_output=True).returncode == 0
    if not conda_exist:
        print("Conda is not installed. Exiting...")
        sys.exit()

    # Ensure this is a new environment and not the base environment
    if os.environ["CONDA_DEFAULT_ENV"] == "base":
        print("Create an environment for this project and activate it. Exiting...")
        sys.exit()

def install_dependencies():
    # Install the webui dependencies
    print('正在使用阿里源安装依赖...\n')
    run_cmd("python -m pip install -r requirements.txt --upgrade -i https://mirrors.aliyun.com/pypi/simple/", assert_success=True, environment=True)

ready_msg = """

一切准备就绪!

- （重要）请确保 gpt_academic/config.py 中已经配置了正确的设置选项.
- 关闭此窗口即可结束本程序, 下次运行时, 仍然是启动`一键本地运行gpt-学术优化.bat` 
- 如需卸载, 删除整个文件夹即可

当您准备完毕后, 请敲回车继续.
"""

def run_model():
    input(ready_msg)
    run_cmd(f"python main.py", environment=True)


if __name__ == "__main__":
    if not os.path.exists('./gpt_academic/main.py'):
        print('从Github下载`https://github.com/guoruqiang/gpt_academic/archive/refs/heads/master.zip`。')
        print('把压缩包中的`gpt_academic-master`文件夹解压到当前路径下。')
        print(f'等待完成 ')


        while True:
            if os.path.exists('./gpt_academic-master/main.py'):
                os.rename('gpt_academic-master', 'gpt_academic')
                break
            input('尚未检测到gpt_academic-master文件夹, 敲回车重试.')

    # Verifies we are in a conda environment
    check_env()

    # If webui has already been installed, skip and run
    if os.path.exists("gpt_academic/"):
        os.chdir("gpt_academic")
        install_dependencies()
        run_model()
